static char rcsid[] = "@(#)$Id: okay_addr.c,v 1.17.8.8 2006/08/26 16:32:44 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.17.8.8 $   $State: Exp $
 *
 *  Reimplemented by: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *                            (was hurtta+elm@ozone.FMI.FI)
 *                or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 *  This code is reimplemented. Original code on Elm 2.4 lib/okay_addr.c 
 *  was following copyright:
 *
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** 

**/

#include "elm_defs.h"


const static unsigned char *cs2us P_((const char *str));
const static unsigned char *cs2us(str)
     const char *str;
{
    return (const unsigned char *)str;
}


DEBUG_VAR(Debug,__FILE__,"addr");

/* returns:   1 if dequoted,
              0 if unmodified
	     -1 if overflow
	     -2 if syntax error
*/

static int strcpy_dequote P_((char * target, const char * source, int size));
static int strcpy_dequote(target,source,size)
     char * target; 
     const char * source; 
     int size;
{
    int r = 0;

    int q = 0;
    const char *s;
    char *t;
    int overflow     = 0;
    int syntax_error = 0;

#define ADD(x) { int X = (x); \
	if (t < target + size - 1) *t++=X; else overflow++; \
	}

    for (s = source, t = target; *s; s++) {
	if (q) {
	    if ('"' == *s) {
		q = !q;
	    } else if ('\\' == *s) {
		s++;
		if (!*s) {
		    syntax_error++;
		}
		ADD(*s);
	    } else {
		ADD(*s);
	    }
	} else {
	    if ('"' == *s) {
		r = 1;
		q = !q;
	    } else {
		if ('\\' == *s) {
		    syntax_error++;
		}
		ADD(*s);
	    }
	}
    }
    *t = '\0';

    if (syntax_error) {
	DPRINT(Debug,2,(&Debug,
			"strcpy_dequote: syntax error: source=%s\n",
			source));		   
	r = -2;
    } else if (overflow) {
	DPRINT(Debug,2,(&Debug,
			"strcpy_dequote: overflow: source=%s\n",
			source));
	
	r = -1;
    }

    return r;

}


static int compare_helper P_((const char *A, const char *pat,
			      int ignore_case));
static int compare_helper(A,pat,ignore_case)
     const char *A; 
     const char *pat;
     int ignore_case;
{
    /* NOTE: Inefficient */

    struct string *SA   = new_string2(ASCII_SET,cs2us(A));
    struct string *Spat = new_string2(ASCII_SET,cs2us(pat));

    int r = string_match(SA,Spat,ignore_case);


    free_string(&Spat);
    free_string(&SA);
    
    return r;
}

#define SPLIT_ADDR_magic        0xEF01

void clear_split_addr(addr)
     struct split_addr *addr;
{
    if (SPLIT_ADDR_magic != (addr)->magic)
	panic("ADDR PANIC",__FILE__,__LINE__,"free_split_addr",
	      "Bad magic number",0);

    if ((addr)->local_part) {
	free((addr)->local_part);
	(addr)->local_part = NULL;
    }
    (addr)->local_part_dequote = 0;

    if ((addr)->local_part_raw) {
	free((addr)->local_part_raw);
	(addr)->local_part_raw = NULL;
    }

    if ((addr)->host_part) {
	free((addr)->host_part);
	(addr)->host_part = NULL;
    }
}


/* Return 1, if address was splitted */
int split_internet_address(result,addr)
     struct split_addr *result;
     const char *addr;
{
    const char * host_part = qstrpbrk((char *)addr,"@");

    result->magic  = 0;


    if (host_part) {
	
	int LEN = host_part - addr +1;
	char * buffer = safe_malloc(LEN);
	char * buffer1 = safe_malloc(LEN);
	int D1;
	
	strncpy(buffer1,addr,host_part - addr);
	buffer1[host_part - addr] = '\0';
	
	D1 = strcpy_dequote(buffer,buffer1,LEN);
       
	DPRINT(Debug,27,(&Debug,
			 "split_internet_address: %s => user part=%s%s, host part=%s\n",
			 addr,buffer,
			 D1 > 0 ? " (dequoted)" : "",
			 host_part+1));


	result->magic = SPLIT_ADDR_magic;
	result->local_part = buffer;
	result->local_part_dequote = D1;
	result->local_part_raw = buffer1;

	result->host_part  = safe_strdup(host_part+1);
	return 1;
    }

    return 0;
}

static int is_pattern P_((char *xx));
static int is_pattern(xx)
     char *xx;
{
    return strpbrk(xx,"*?") != NULL;
}

int match_splitted_addr(addr,pattern)
     struct split_addr *addr;
     struct split_addr *pattern;
{

    if (SPLIT_ADDR_magic != addr->magic)
	panic("ADDR PANIC",__FILE__,__LINE__,"match_splitted_addr",
	      "Bad magic number (addr)",0);
    if (SPLIT_ADDR_magic != pattern->magic)
	panic("ADDR PANIC",__FILE__,__LINE__,"match_splitted_addr",
	      "Bad magic number (pattern)",0);


    if (0 == strcmp(addr->local_part,
		    pattern->local_part) &&
	0 == istrcmp(addr->host_part,
		     pattern->host_part)) {
			
	return 1;
    }

    if (is_pattern(pattern->local_part) ||
	is_pattern(pattern->host_part)) {

	if (compare_helper(addr->local_part,
			   pattern->local_part,0) 
	    &&
	    compare_helper(addr->host_part,
			   pattern->host_part,1)) {

	    return 1;
	}
    }

    return 0;
}

int match_local_addr(addr,pattern)
     char *addr;
     char *pattern;
{

    if (0 == strcmp(addr,pattern)) {
	return 1;
    }

    if (is_pattern(pattern)) {
	if (compare_helper(addr,pattern,0)) {
	    return 1;
	}
    }

    return 0;
}

int addr_is_user(addr)
     char *addr;
{
    char **alternatives;
    char our_address[SLEN]; 
    char deqbuffer[SLEN];

    struct split_addr  splitted_addr;

    int D = strcpy_dequote(deqbuffer, addr, sizeof deqbuffer);

    if (D != 0) {
	DPRINT(Debug,25,(&Debug,
			 "addr_is_user(\"%s\"): Dequotation: %s\n",
			 addr,deqbuffer));
    }

    if (0 == strcmp(username, deqbuffer)) {
	DPRINT(Debug,25,(&Debug,
			 "addr_is_user(\"%s\")=TRUE: %s matches username\n",
			 D > 0 ? "(dequoted)" : "",
			 addr));
	return TRUE;
    }
	
          
    /* UUCP hostnames are case sensitive */
    elm_sfprintf(our_address, sizeof our_address,
		 FRM("%s!%s"), hostname, username);

    if (0 == strcmp(our_address, deqbuffer)) {
	DPRINT(Debug,25,(&Debug,				    
			 "addr_is_user(\"%s\")=TRUE: %s matches UUCP address\n",
			 D > 0 ? "(dequoted)" : "",
			 addr));
	return TRUE;
    }

    elm_sfprintf(our_address, sizeof our_address,
		 FRM("%s!%s"), hostfullname, username);

    if (0 == strcmp(our_address, deqbuffer)) {
	DPRINT(Debug,25,(&Debug,
			 "addr_is_user(\"%s\")=TRUE: %s matches UUCP address\n",
			 D > 0 ? "(dequoted)" : "",
			 addr));
	return TRUE;
    }
    
    /* Internet hostnames are case insensitive */

    if (split_internet_address(&splitted_addr,addr)) {

       
	DPRINT(Debug,25,(&Debug,
			 "addr_is_user(\"%s\"): user part=%s%s, host part=%s\n",
			 addr,
			 splitted_addr.local_part,
			 splitted_addr.local_part_dequote > 0 ? 
			 " (dequoted)" : "",
			 splitted_addr.host_part));
    
	if (0 == strcmp(splitted_addr.local_part,username)) {

	    if (0 == istrcmp(hostname,splitted_addr.host_part)) {
		DPRINT(Debug,25,(&Debug,
				 "addr_is_user(\"%s\")=TRUE: matches internet address\n",
				 addr));

		clear_split_addr(&splitted_addr);

		return(TRUE);
	    }
    
	    if (0 == istrcmp(hostfullname,splitted_addr.host_part)) {
		DPRINT(Debug,25,(&Debug,
				 "addr_is_user(\"%s\")=TRUE: matches internet address\n",
				 addr));		
			
		clear_split_addr(&splitted_addr);

		return(TRUE);
	    }
	}

	alternatives = give_dt_path_as_elems(&alternative_addresses,
					     "alternatives");
	if (alternatives) {

	    int i;

	    for (i = 0; alternatives[i]; i++) {

		struct split_addr  splitted_alternative;
		
		if (split_internet_address(&splitted_alternative,
					   alternatives[i])) {
		  
		    DPRINT(Debug,26,(&Debug,
				     "addr_is_user: %s: user part=%s%s, host part=%s\n",
				     alternatives[i],
				     splitted_alternative.local_part,
				     splitted_alternative.local_part_dequote > 0 ? 
				     " (dequoted)" : "",
				     splitted_alternative.host_part));


		    if (match_splitted_addr(& splitted_addr,
					    & splitted_alternative)) {

			DPRINT(Debug,25,(&Debug,
					 "addr_is_user(\"%s\")=TRUE: matches alternative\n",
					 addr));		
			
			clear_split_addr(&splitted_alternative);
			clear_split_addr(&splitted_addr);
			
			return(TRUE);		    
		    }

		    clear_split_addr(&splitted_alternative);
		}
	    }
	}

	clear_split_addr(&splitted_addr);
    }

    alternatives = give_dt_path_as_elems(&alternative_addresses,
					 "alternatives");    
    if (alternatives) {

	int i;
	
	for (i = 0; alternatives[i]; i++) {
	    
	    if (match_local_addr(addr,alternatives[i])) {
		DPRINT(Debug,25,(&Debug,				   
				 "addr_is_user(\"%s\")=TRUE: matches alternative address\n",
				 addr));		
		return(TRUE);
	    }

	}
    }
    DPRINT(Debug,25,(&Debug,
		     "addr_is_user(\"%s\")=FALSE\n",addr));
    return FALSE;
}

int okay_address_l(address, return_address)
     const struct addr_item *address, *return_address;
{
    
    /** This routine checks to ensure that the address we just got
	from the "To:" or "Cc:" line isn't us AND isn't the person	
	who sent the message.  Returns true iff neither is the case **/
    
    const struct addr_item *ptr;
    
    for (ptr = return_address; ptr && ptr -> addr && ptr->fullname; ptr++) {
	if (0 == strcmp(ptr -> addr, address -> addr)) {
	    DPRINT(Debug,25,(&Debug,
			     "okay_address_l=FALSE:\n"));
	    DPRINT(Debug,25,(&Debug,
			     "     address{addr=%s,fullname=%S} same as return_address {addr=%s,fullname=%S}\n",
			     address -> addr, address -> fullname,
			     ptr -> addr, ptr -> fullname));
	    return(FALSE);
	}
    }

    if (addr_is_user(address -> addr)) {
	DPRINT(Debug,25,(&Debug,
		   "okay_address_l=FALSE:\n"));
	DPRINT(Debug,25,(&Debug,
			 "     address{addr=%s,fullname=%S} is user's address\n",
			 address -> addr, address -> fullname));	
	return(FALSE);
    }
    
    DPRINT(Debug,25,(&Debug,
		     "okay_address_l=TRUE:\n"));
    DPRINT(Debug,25,(&Debug,
		     "     address{addr=%s,fullname=%S} is not our address.\n",
		     address -> addr, address -> fullname)); 
    return(TRUE);
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
